@@ -55,7 +55,7 @@ use Geo::GDAL;
 use Data::Dumper;
 use XML::LibXML::PrettyPrint;
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 our %gml_geometry_type = (
     Envelope => 1,
@@ -79,6 +79,28 @@ our %spatial2op = (
     Contains => 1
     ); # there are more in PostGIS
 
+our %temporal_operators = (
+    After => 1,
+    Before => 1,
+    Begins => 1,
+    BegunBy => 1,
+    TContains => 1,
+    During => 1,
+    TEquals => 1,
+    TOverlaps => 1,
+    Meets => 1,
+    OverlappedBy => 1,
+    MetBy => 1,
+    EndedBy => 1
+    );
+
+my %functions = (
+    abs => ['xs:int', [[int => 'xs:int']]],
+    acos => ['xs:double', [[value => 'xs:double']]],
+    Area => ['xs:double', [[geometry => 'gml:AbstractGeometryType']]],
+    within => ['xs:boolean', [[geometry => 'gml:AbstractGeometryType'], [geometry => 'gml:AbstractGeometryType']]]
+    ); # lots missing
+
 # GDAL and PostgreSQL data type to XML data type
 our %type_map = (
     geometry => "gml:GeometryPropertyType",
@@ -102,6 +124,15 @@ our %type_map = (
     boolean => "xs:boolean",
     );
 
+# Value in request => GDAL GML creation option
+our %OutputFormats = (
+    'XMLSCHEMA' => 'GML2',
+    'text/xml; subtype=gml/2.1.2' => 'GML2',
+    'text/xml; subtype=gml/3.1.1' => 'GML3',
+    'text/xml; subtype=gml/3.2' => 'GML3.2',
+    'GML3Deegree' => 'GML3Deegree',
+    );
+
 =pod
 
 =head3 process_request
@@ -267,9 +298,9 @@ sub OperationsMetadata  {
                       {AcceptVersions => \@versions}, 
                       {AcceptFormats => ['text/xml']}]);
     $self->Operation($writer, 'DescribeFeatureType', 
-                     [{outputFormat => ['XMLSCHEMA','text/xml; subtype=gml/2.1.2','text/xml; subtype=gml/3.1.1']}]);
+                     [{outputFormat => [sort keys %OutputFormats]}]);
     $self->Operation($writer, 'GetFeature',
-                     [{resultType => ['results']}, {outputFormat => ['text/xml; subtype=gml/3.1.1']}]);
+                     [{resultType => ['results']}, {outputFormat => [sort keys %OutputFormats]}]);
     $self->Operation($writer, 'Transaction',
                      [{inputFormat => ['text/xml; subtype=gml/3.1.1']}, 
                       {idgen => ['GenerateNew','UseExisting','ReplaceDuplicate']},
@@ -336,13 +367,17 @@ sub FeatureTypeList  {
     my $list_layer = sub {
         my ($type) = @_;
         $self->get_layer($type); # test only, now opens.. could be just name
+        my @formats;
+        for my $format (sort keys %OutputFormats) {
+            push @formats, ['wfs:Format', $format];
+        }
         my @FeatureType = (
             ['wfs:Name', $type->{Name}],
             ['wfs:Title', $type->{Title}],
             ['wfs:Abstract', $type->{Abstract}],
             ['wfs:DefaultSRS', $type->{DefaultSRS}],
             ['wfs:OtherSRS', 'EPSG:3857'],
-            ['wfs:OutputFormats', ['wfs:Format', 'text/xml; subtype=gml/3.1.1']]);
+            ['wfs:OutputFormats', \@formats]);
         push @FeatureType, ['ows:WGS84BoundingBox', {dimensions=>2}, 
                             [['ows:LowerCorner',$type->{LowerCorner}],
                              ['ows:UpperCorner',$type->{UpperCorner}]]]
@@ -381,7 +416,27 @@ sub Filter_Capabilities  {
     my $ns = $self->{version} eq '2.0.0' ? 'fes' : 'ogc';
     $writer->open_element($ns.':Filter_Capabilities');
 
-    # fes:Conformance missing
+    # Conformance
+    my %Constraints = ( 
+        ImplementsQuery => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
+        ImplementsAdHocQuery => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
+        ImplementsFunctions => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
+        ImplementsMinStandardFilter => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
+        ImplementsStandardFilter => [['ows:NoValues'], ['ows:DefaultValue' => 'FALSE']],
+        ImplementsMinSpatialFilter => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
+        ImplementsSpatialFilter => [['ows:NoValues'], ['ows:DefaultValue' => 'FALSE']],
+        ImplementsMinTemporalFilter => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
+        ImplementsTemporalFilter => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
+        ImplementsVersionNav => [['ows:NoValues'], ['ows:DefaultValue' => 'FALSE']],
+        ImplementsSorting => [['ows:AllowedValues' => [['ows:Value' => 'ASC'], ['ows:Value' => 'DESC']]], 
+                              ['ows:DefaultValue' => 'ASC']],
+        ImplementsExtendedOperators => [['ows:NoValues'], ['ows:DefaultValue' => 'FALSE']],
+        );
+    my @c;
+    for my $key (keys %Constraints) {
+        push @c, [$ns.':Constraint', {name=>$key}, $Constraints{$key}];
+    }
+    $writer->element($ns.':Conformance', \@c);
 
     my @ids;
     if ($ns eq 'ogc') {
@@ -421,10 +476,39 @@ sub Filter_Capabilities  {
                 [[$ns.':GeometryOperands', \@operands],
                  [$ns.':SpatialOperators', \@operators]]);
 
-    # fes:Temporal_Capabilities missing
+    # Temporal_Capabilities
 
-    # fes:Functions missing
+    @operands = ();
+    for my $o (qw/TimeInstant TimePeriod/) {
+        if ($ns eq 'ogc') {
+            push @operands, [$ns.':GeometryOperand', 'gml:'.$o];
+        } else {
+            push @operands, [$ns.':GeometryOperand', { name => 'gml:'.$o }];
+        }
+    }
+    @operators = ();
+    @op = keys %temporal_operators;
+    for my $o (@op) {
+        push @operators, [$ns.':TemporalOperator', { name => $o }];
+    }
+    $writer->element($ns.':Temporal_Capabilities', 
+                     [[$ns.':TemporalOperands', \@operands],
+                      [$ns.':TemporalOperators', \@operators]]);
 
+    # Functions
+    
+    my @functions;
+    for my $f (sort keys %functions) {
+        my @args = ();
+        my $args = $functions{$f}[1];
+        for my $arg (@$args) {
+            push @args, [$ns.':Argument', { name => $arg->[0]}, [$ns.':Type' => $arg->[1]]];
+        }
+        push @functions, [$ns.':Function', { name => $f }, 
+                          [[$ns.':Returns' => $functions{$f}[0]] ,[$ns.':Arguments' => \@args]]];
+    }
+    $writer->element($ns.':Functions', \@functions);
+    
     $writer->close_element;
 }
 
@@ -572,7 +656,7 @@ sub GetFeature {
     elsif ($type && $type->{Table}) {
 
         my $filter = filter2sql($query->{filter}, $type) // '';
-        my $epsg = $query->{EPSG} // '';
+        my $epsg = $query->{EPSG};
 
         # pseudo_credentials: these fields are required to be in the filter and they are not included as attributes
         my ($pseudo_credentials, @pseudo_credentials) = pseudo_credentials($type);
@@ -581,11 +665,12 @@ sub GetFeature {
             my $pat1 = "\\(\\(\"$pseudo_credentials[0]\" = '.+?'\\) AND \\(\"$pseudo_credentials[1]\" = '.+?'\\)\\)";
             my $pat2 = "\\(\\(\"$pseudo_credentials[1]\" = '.+?'\\) AND \\(\"$pseudo_credentials[0]\" = '.+?'\\)\\)";
             my $n = join(' and ', @pseudo_credentials);
-            $self->error({ exceptionCode => 'InvalidParameterValue',
-                           locator => 'filter',
-                           ExceptionText => "Not authorized. Please provide '$n' in filter." })
-                unless $filter and ($filter =~ /$pat1/ or $filter =~ /$pat2/);
-            return;
+            unless ($filter and ($filter =~ /$pat1/ or $filter =~ /$pat2/)) {
+                $self->error({ exceptionCode => 'InvalidParameterValue',
+                               locator => 'filter',
+                               ExceptionText => "Not authorized. Please provide '$n' in filter." });
+                return;
+            }
         }
         
         my @cols;
@@ -601,7 +686,7 @@ sub GetFeature {
             # only one geometry property in the output
             next if $type->{Schema}{$col}{out_type} eq 'gml:GeometryPropertyType' && not $is_geometry;
 
-            if ($epsg and $is_geometry) {
+            if (defined $epsg and $is_geometry) {
                 push @cols, "st_transform(\"$col\",$epsg) as \"$name\"";
             } else {
                 $filter =~ s/$name/$col/g if $filter;
@@ -620,7 +705,10 @@ sub GetFeature {
         $sql .= " AND (\"$geom\" && ST_MakeEnvelope(".join(",", @{$query->{BBOX}})."))" if $query->{BBOX};
 
         print STDERR "$sql\n" if $self->{debug} > 2;
-        $layer = $self->{DataSource}->ExecuteSQL($sql);
+        eval {
+            $layer = $self->{DataSource}->ExecuteSQL($sql);
+        };
+        print STDERR $@ if $@;
     }
 
     unless ($layer) {
@@ -636,7 +724,9 @@ sub GetFeature {
     # use "TARGET_NAMESPACE": "http://ogr.maptools.org/", "PREFIX": "ogr", in config or type section
     my $ns = $self->{config}{TARGET_NAMESPACE} // $type->{TARGET_NAMESPACE} // 'http://www.opengis.net/wfs';
     my $prefix = $self->{config}{PREFIX} // $type->{PREFIX} // 'wfs';
-    my $format = $self->{config}{FORMAT} // 'GML3.2';
+    my $format = $self->{request}{outputformat} // $self->{config}{FORMAT} // 'GML3.2';
+    $format = $OutputFormats{$format} if $OutputFormats{$format};
+    print STDERR "Output format (for GDAL creation): $format\n" if $self->{debug} > 2;
 
     my $vsi = '/vsistdout/';
     my $gml;
@@ -840,6 +930,7 @@ sub parse_request {
         $self->{request} = {
             request => $self->{parameters}{request},
             version => $self->{parameters}{version},
+            outputformat => $self->{parameters}{outputformat},
             queries => [
                 {
                     typename => fallback($self->{parameters}{typename}, $self->{parameters}{typenames}),
@@ -850,6 +941,7 @@ sub parse_request {
                 }
                 ]
         };
+        $self->{request}{outputformat} = 'GML2' if $self->{request}{version} && $self->{request}{version} eq '1.0.0';
     }
 
     my %defaults = (
@@ -881,12 +973,13 @@ sub open_datasource {
 
 sub feature_type {
     my ($self, $type_name) = @_;
+    my ($ns, $name) = parse_name($type_name);
     for my $type (@{$self->{config}{FeatureTypeList}}) {
         my $ret;
         eval {
             if ($self->open_datasource($type)) {
                 if (exists $type->{Layer}) {
-                    if ($type->{Name} eq $type_name) {
+                    if ($type->{Name} eq $name) {
                         $ret = clone($type);
                         my $schema = $self->get_layer($type)->GetSchema;
                         my %geometry_types = map {$_ => 1} Geo::OGR::GeometryTypes();
@@ -909,7 +1002,7 @@ sub feature_type {
                 else {
                     # open policy: announce all in a data source except those denied
                     for my $t ($self->feature_types_in_data_source($type)) {
-                        if ($t->{Name} eq $type_name) {
+                        if ($t->{Name} eq $name) {
                             $ret = $t;
                         }
                     }
@@ -980,6 +1073,7 @@ sub feature_types_in_data_source {
             push @geometry_columns, $n if $data->{TYPE_NAME} eq 'geometry';            
         }
         for my $geom (@geometry_columns) {
+            print STDERR "Testing \"$table.$geom\": " if $self->{debug} > 2;
             my $sql = "select auth_name,auth_srid ".
                 "from \"$table\" ".
                 "join spatial_ref_sys on spatial_ref_sys.srid=st_srid(\"$geom\") ".
@@ -1014,8 +1108,14 @@ sub feature_types_in_data_source {
             $shortname =~ s/[ÅÖ]/O/g;
             $shortname =~ s/Ä/A/g;
             my $name = $prefix.'.'.$shortname;
-            next if $type->{allow} and (!$type->{allow}{$shortname} or !$type->{allow}{$name});
-            next if $type->{deny} and ($type->{deny}{$shortname} or $type->{deny}{$name});
+            if ($type->{allow} and !($type->{allow}{$shortname} or $type->{allow}{$name})) {
+                print STDERR "not allowed.\n" if $self->{debug} > 2;
+                next;
+            }
+            if ($type->{deny} and ($type->{deny}{$shortname} or $type->{deny}{$name})) {
+                print STDERR "denied.\n" if $self->{debug} > 2;
+                next;
+            }
 
             my $feature_type = {
                 Title => "$table($geom)",
@@ -1036,6 +1136,17 @@ sub feature_types_in_data_source {
             for my $key (keys %$type) {
                 $feature_type->{$key} //= $type->{$key} unless ref $key;
             }
+            my ($h, @c) = pseudo_credentials($feature_type);
+            my $ok = 1;
+            for my $c (@c) {
+                unless ($feature_type->{Schema}{$c}) {
+                    print STDERR "pseudo credential column '$c' not in schema.\n" if $self->{debug} > 2;
+                    $ok = 0;
+                    next;
+                }
+            }
+            next unless $ok;
+            print STDERR "added.\n" if $self->{debug} > 2;
             push @feature_types, $feature_type;
         }
     }
@@ -1051,10 +1162,8 @@ sub get_layer {
 # return WFS request in a hash
 # this function is written according to WFS 1.1.0 / 2.0.0
 sub ogc_request {
-    my($node) = @_;
-    my $ns = '';
-    my $name = $node->nodeName;
-    ($ns, $name) = split /:/, $name if $name =~ /:/;
+    my ($node) = @_;
+    my ($ns, $name) = parse_tag($node);
     if ($name eq 'GetCapabilities') {
         return { service => $node->getAttribute('service'), 
                  request => $name,
@@ -1115,8 +1224,8 @@ sub ogc_request {
             my $b = $node->getAttribute($a);
             $query->{lc($key)} = $b if $b;
         }
-        if (defined $query->{srsName}) {
-            ($query->{EPSG}) = $query->{srsName} =~ /EPSG:(\d+)/;
+        if (defined $query->{srsname}) {
+            $query->{EPSG} = get_integer($query->{srsname});
         }
         for my $property ($node->getChildrenByTagNameNS('*', 'PropertyName')) {
             $query->{properties} = {} unless $query->{properties};
@@ -1332,10 +1441,16 @@ sub filter2sql {
 
     } elsif ($name eq 'BBOX') {
         $node = $node->firstChild;
-        my $property = filter2sql($node, $type);
-        $node = $node->nextSibling;
-        my $envelope = filter2sql($node, $type);
-        return "($property && $envelope)";
+        my ($ns, $n) = parse_tag($node);
+        if ($n eq 'Envelope') {
+            my $envelope = filter2sql($node, $type);
+            return "(GeometryColumn && $envelope)";
+        } else {
+            my $property = filter2sql($node, $type);
+            $node = $node->nextSibling;
+            my $envelope = filter2sql($node, $type);
+            return "($property && $envelope)";
+        }
 
     } elsif ($spatial2op{$name}) {
         $node = $node->firstChild;
@@ -1353,6 +1468,10 @@ sub filter2sql {
         my $dist = filter2sql($node, $type);
         return "ST_$name($geom1, $geom2, $dist)";
 
+    # temporal filter missing
+
+    # functions missing
+
     } elsif ($name eq 'Filter') {
         my %id = ( ResourceId => 1,
                    FeatureId => 1,
@@ -1374,10 +1493,8 @@ sub filter2sql {
 
 sub node2sql {
     my ($node) = @_;
-    my $srs = $node->getAttribute('srsName');
-    ($srs) = $srs =~ /EPSG:(\d+)/ if $srs;
-    $srs = ",$srs" if $srs;
-    $srs //= '';
+    my $srs = get_integer($node->getAttribute('srsName')) // '';
+    $srs = ",$srs" if $srs ne '';
     my ($ns, $name) = parse_tag($node);
     my $wkt;
 
@@ -1460,8 +1577,12 @@ sub error {
 
 sub parse_tag {
     my $node = shift;
+    return parse_name($node->nodeName);
+}
+
+sub parse_name {
+    my $name = shift;
     my $ns = '';
-    my $name = $node->nodeName;
     ($ns, $name) = split /:/, $name if $name =~ /:/;
     return ($ns, $name);
 }
